/// <reference path="./modules.d.ts" />
import './redirect';
import { uniq } from 'lodash';
import type { Merge, SetRequired } from 'type-fest';
import type { Rule } from 'eslint';
import type { ESLintUtils } from '@typescript-eslint/utils';
import type {
  ESLintConfig,
  Extends,
  KnownExtends,
  Plugin,
  Rules,
  Settings,
} from '@aet/eslint-define-config';

import { importTypeScript } from './presets/typescript';
import { unicorn } from './presets/unicorn';
import { eslintRules } from './presets/eslint';
import { local } from './presets/local';
import { error, warn, off } from './constants';
import { checkEnv } from './env';

export { graphql } from './presets/graphql';
export { jsdoc } from './presets/jsdoc';
export { storybook } from './presets/misc';
export { react, reactRefresh } from './presets/react';
export { tailwind } from './presets/tailwind';

export { error, warn, off };

declare global {
  interface Array<T> {
    filter(
      predicate: BooleanConstructor,
    ): Exclude<T, null | undefined | false | '' | 0>[];
  }
}

const unique = (...arr: (false | undefined | string | string[])[]): string[] => [
  ...new Set(arr.flat(1).filter(Boolean)),
];

const ensureArray = <T>(value?: T | T[]): T[] =>
  value == null ? [] : Array.isArray(value) ? value : [value];

type RuleLevel = 'error' | 'warn' | 'off' | 0 | 1 | 2;
type RuleEntry<Options> = RuleLevel | [RuleLevel, Partial<Options>];

export interface LocalRuleOptions {
  /** Bans import from the specifier '.' and '..' and replaces it with '.+/index' */
  'rules/no-import-dot': RuleEntry<unknown>;
  /**
   * Enforce template literal expressions to be of `string` type
   * @see [restrict-template-expressions](https://typescript-eslint.io/rules/restrict-template-expressions)
   */
  'rules/restrict-template-expressions': RuleEntry<{ allow: string[] }>;
  /** Ban assignment of empty object literals `{}` and replace them with `Object.create(null)` */
  'rules/no-empty-object-literal': RuleEntry<unknown>;
}

export type RuleOptions = Rules & Partial<LocalRuleOptions>;

export interface CustomRule {
  rule: () => Promise<{
    default: Rule.RuleModule | ESLintUtils.RuleModule<string, unknown[]>;
  }>;
  options?: RuleLevel;
}

export type Middleware = (config: MiddlewareConfig, helpers: MiddlewareFunctions) => void;

/**
 * ESLint Configuration.
 * @see [ESLint Configuration](https://eslint.org/docs/latest/user-guide/configuring/)
 */
export type InputConfig = Omit<ESLintConfig, 'rules'> & {
  /**
   * Rules.
   * @see [Rules](https://eslint.org/docs/latest/user-guide/configuring/rules)
   */
  rules?: Partial<RuleOptions>;

  /**
   * Glob pattern to find paths to custom rule files in JavaScript or TypeScript.
   * Note this must be a string literal or an array of string literals since
   * this is statically analyzed.
   */
  customRuleFiles?: string | string[];

  /**
   * Automatically detect project types, dependencies and deduct the plugins.
   */
  auto?: boolean;
};

type OptionalObjectKey<T> = Exclude<
  {
    [Key in keyof T]: undefined | any[] extends T[Key]
      ? Key
      : undefined | Record<any, any> extends T[Key]
        ? Key
        : never;
  }[keyof T],
  undefined
>;

type MiddlewareConfig = Merge<
  SetRequired<ESLintConfig, OptionalObjectKey<ESLintConfig>>,
  { extends: KnownExtends[] }
>;

interface MiddlewareFunctions {
  addRules(rules: Partial<RuleOptions>): void;
  addSettings(settings: Partial<Settings>): void;
}

/**
 * Returns a ESLint config object.
 *
 * By default, it includes `["@typescript-eslint", "import-x", "prettier", "unicorn"]` configs.
 * Additional bundled plugins include:
 *
 * 1. [`react`](https://github.com/jsx-eslint/eslint-plugin-react#list-of-supported-rules)
 *    (automatically enables
 *    [`react-hooks`](https://github.com/facebook/react/tree/main/packages/eslint-plugin-react-hooks))
 * 2. [`react-refresh`](https://github.com/ArnaudBarre/eslint-plugin-react-refresh)
 * 3. [`jsx-a11y`](https://github.com/jsx-eslint/eslint-plugin-jsx-a11y#supported-rules)
 * 4. [`unicorn`](https://github.com/sindresorhus/eslint-plugin-unicorn#rules)
 * 5. [`n`](https://github.com/eslint-community/eslint-plugin-n#-rules) (Node.js specific,
 *    requires `minimatch`)
 * 6. [`jsdoc`](https://github.com/gajus/eslint-plugin-jsdoc#rules)
 *
 * Non bundled:
 * 1. [`graphql`](https://the-guild.dev/graphql/eslint/rules)
 */
export function extendConfig(
  of: InputConfig & {
    middlewares: Middleware[];
  } = {
    middlewares: [],
  },
): ESLintConfig {
  const {
    auto = true,
    plugins: _plugins = [],
    settings = {},
    rules,
    extends: _extends,
    overrides,
    customRuleFiles,
    parserOptions,
    middlewares: _middlewares = [],
    // @ts-expect-error
    localRules: _,
    ...rest
  } = of;

  let plugins: Plugin[] = [..._plugins];
  let extend: Extends[] = ensureArray(_extends);

  const middlewares: Middleware[] = uniq([
    importTypeScript,
    unicorn,
    local,
    ...(auto ? checkEnv() : []),
    ..._middlewares,
  ]);

  const result: MiddlewareConfig = {
    root: true,
    plugins: unique('rules', plugins),
    env: { node: true, browser: true, es2023: true },
    reportUnusedDisableDirectives: true,
    parserOptions: {
      project: true,
      ...parserOptions,
    },
    ignorePatterns: [],
    globals: {},
    extends: ['eslint:recommended', 'prettier', ...(extend as string[])],
    settings,
    overrides: [
      { files: ['repl.ts', 'scripts/**/*.ts'], rules: { 'no-console': off } },
      ...(overrides ?? []),
    ],
    rules: { ...eslintRules, ...rules },
    ...rest,
  };

  const functions: MiddlewareFunctions = {
    addRules(newRules) {
      Object.assign(result.rules, newRules);
    },
    addSettings(newSettings) {
      Object.assign(result.settings, newSettings);
    },
  };

  for (const fn of middlewares) {
    fn(result, functions);
  }

  result.plugins = unique(result.plugins);
  result.extends = unique(result.extends);

  return result;
}
